//--------------------------------------------------------------------------------------------------
//	PROGRAM : I2C + PCF8583 + EEPROM routines
//	VERSION : 1.0
//	AUTHOR  : Bc. Pazour Zbynìk
//	DATE	: 07.09.2009
//--------------------------------------------------------------------------------------------------
#ifndef	 _i2c_h
#define	 _i2c_h

#define MENU    PINB3
#define ESC 	PINB2
#define UP 		PINB1
#define DOWN 	PINB0



void kontrola (unsigned short STAV);
void twi_start (void);
void hotovo (void);
void vysli (unsigned short ADR);
void cti_data_ACK (void);
void cti_data_NACK (void);
void twi_stop (void);
void i2c_init (void);
void ERROR (void);
void zobraz_cas (void);
void nacti_RTC (void);
void zapis_rok (void);
void nastav_cas (void);
void nastav_citac (char citac);
void nul_citac (void);
#endif
